package octoshape.android.stream;

import android.net.Uri;
import android.util.Log;

import com.google.android.gms.cast.MediaInfo;

import octoshape.android.CastApplication;

import org.json.JSONException;
import org.json.JSONObject;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;

import octoshape.osa2.Problem;
import octoshape.osa2.android.StreamPlayer;
import octoshape.osa2.android.listeners.MediaPlayerListener;
import octoshape.osa2.android.listeners.StreamPlayerListener;
import octoshape.osa2.listeners.ProblemListener;

/**
 * A wrapper for the {@linkplain octoshape.osa2.android.StreamPlayer}
 *
 * Created by dota on 18-02-2015.
 */
public class StreamPlayerWrapper implements ProblemListener {

    public final static String TAG = "StreamPlayerWrapper";
    public static final String IS_LIVE = "isLive";
    public static final String SEEK_TYPE = "seekType";
    public static final String DURATION = "duration";
    public static final String SEEK_OFFSET = "seekOffset";
    public static final String CC_STRING =  "ccs";
    private MediaInfo mediaInfo;

    /**
     * The wrapped Stream Player
     */
    private StreamPlayer streamPlayer;

    /**
     * octoshape:// link
     */
    private String stream;

    /**
     * StreamPlayer listeners
     */
    private IStreamPlayerWrapperListener listener;

    /**
     * Map from the video URL to the associated MediaPlayerListener
     */
    private MediaPlayerListener mediaPlayerListener;

    /**
     * The playback state for each video URL
     */
    private StreamManager.PlaybackState streamPlaybackState;

    /**
     * Is the stream lived?
     */
    private boolean isLive;

    /**
     * Has a max latency?
     */
    private long duration = 0;

    /**
     * Seek type (OSA, NATIVE, NONE)
     */
    private SeekType seekType;

    StreamPlayerWrapper(StreamPlayer sp, String stream, MediaInfo mediaInfo) {
        this.streamPlayer = sp;
        this.mediaInfo = mediaInfo;
        this.stream = stream;
        configureStreamPlayer();
    }


    public void setListener(IStreamPlayerWrapperListener listener) {
        this.listener = listener;

        Log.d(TAG, "Added new listener to " + stream);
    }

    public void removeListener(IStreamPlayerWrapperListener listener) {
        this.listener = null;

        Log.d(TAG, "Removed listener to " + stream);
    }

    /**
     * Configures the stream player and adds the listeners.
     */
    private void configureStreamPlayer() {

    	// required for casting to external devices
    	streamPlayer.setOctolinkOption("airplay", "true");
        streamPlayer.setOctolinkOption("ccs", "false");
    	
        streamPlayer.setProblemListener(this);
        streamPlayer.setListener(new StreamPlayerListener() {

                                     /**
                                      * Receiving new URL from the streamplayer object either due to
                                      * requesting playback, seeking or experiencing a bitrate/resolution
                                      * changes requiring the re-initialization of the Player.
                                      *
                                      * @param url        to be passed to the media player
                                      * @param seekOffset offset we have seek to in milliseconds
                                      */
                                     @Override
                                     public void gotUrl(String url, long seekOffset, MediaPlayerListener mpl) {
                                         mediaPlayerListener = mpl;

                                         if (url == null || url.isEmpty()) {
                                             return;
                                         }

                                         if (mpl == null) {
                                             Log.e(TAG, "Got URL without any Media Player Listener " + stream + "\t" + url);
                                         } else {
                                             Log.d(TAG, "Got Media Player Listener for " + stream + "\t" + url);
                                         }

                                         // parse query options
                                         Uri mediaUrl = Uri.parse(url);

                                         HashMap<String, String> queryOptions = new HashMap<String, String>();
                                         for (String key : mediaUrl.getQueryParameterNames())
                                             queryOptions.put(key, mediaUrl.getQueryParameter(key));

                                         queryOptions.put(IS_LIVE, Boolean.toString(isLive));
                                         queryOptions.put(SEEK_TYPE, Integer.toString(seekType.value));
                                         queryOptions.put(DURATION, Long.toString(duration));
                                         queryOptions.put(SEEK_OFFSET, Long.toString(seekOffset));
                                         queryOptions.put(StreamManager.STREAM_ADDR, stream);

                                         // remove query options from url
                                         mediaUrl = mediaUrl.buildUpon().clearQuery().build();

                                         int streamType;
//                                         if (seekType == SeekType.NATIVE_SEEK) {
//                                             streamType = MediaInfo.STREAM_TYPE_BUFFERED;
//                                         } else {
//                                             streamType = MediaInfo.STREAM_TYPE_LIVE;
//                                         }

                                         if (seekType != SeekType.NO_SEEK && !isLive) {
                                             streamType = MediaInfo.STREAM_TYPE_BUFFERED;
                                         } else {
                                             streamType = MediaInfo.STREAM_TYPE_LIVE;
                                         }

                                         Log.d(TAG, "Stream parameters: " + new JSONObject(queryOptions).toString());

                                         // assemble MediaInfo with actual url and query options as JSON
                                         mediaInfo = new MediaInfo.Builder(mediaUrl.toString()).setContentType(mediaInfo.getContentType())
                                                 .setStreamType(streamType)
                                                 .setStreamDuration(duration)
                                                 .setCustomData(new JSONObject(queryOptions))
                                                 .setMetadata(mediaInfo.getMetadata())
                                                 .build();


                                         if (listener != null) {
                                             listener.streamConnected(mediaInfo);
                                         } else {
                                             Log.d(TAG, "No listener for " + stream);
                                         }
                                         CastApplication.getPlayerManager().updatedMediaInfo(stream, mediaInfo);
                                     }

                                     /**
                                      * Resets an on-demand file duration previously reported in
                                      * resolvedOsaSeek(..) method
                                      */
                                     @Override
                                     public void gotNewOnDemandStreamDuration(long duration) {
                                         Log.d(TAG, "Got new on demand stream duration! + duration");
                                         StreamPlayerWrapper.this.duration = duration;

                                         JSONObject customData = mediaInfo.getCustomData();

                                         if (customData != null) {

                                             try {
                                                 customData.put(DURATION, Long.toString(duration));
                                             } catch (JSONException e) {
                                                 e.printStackTrace();
                                             }


                                             // replace the custom data with the new one
                                             mediaInfo = new MediaInfo.Builder(mediaInfo.getContentId()).setContentType(mediaInfo.getContentType())
                                                     .setStreamType(mediaInfo.getStreamType())
                                                     .setCustomData(customData)
                                                     .setMetadata(mediaInfo.getMetadata())
                                                     .build();

                                             CastApplication.getPlayerManager().updateDuration(stream, duration, mediaInfo);
                                         }
                                     }

                                     /**
                                      * Called if the stream is seekable using the media player's own
                                      * native seeking functionality (e.g., the Android MediaPlayer does
                                      * the seeking for us).
                                      */
                                     @Override
                                     public void resolvedNativeSeek(boolean isLive, String playerId) {
                                         StreamPlayerWrapper.this.isLive = isLive;
                                         seekType = SeekType.NATIVE_SEEK;

                                         Log.d(TAG, "Stream: " + stream + " is NATIVE_SEEK");
                                     }

                                     /**
                                      * Called if it is not possible to seek in the stream.
                                      */
                                     @Override
                                     public void resolvedNoSeek(boolean isLive, String playerId) {
                                         StreamPlayerWrapper.this.isLive = isLive;
                                         seekType = SeekType.NO_SEEK;

                                         Log.d(TAG, "Stream: " + stream + " is NO_SEEK");
                                     }


                                     /**
                                      * Called when stream support OsaSeek / DVR
                                      */
                                     @Override
                                     public void resolvedOsaSeek(boolean isLive, long duration, String playerId) {
                                         StreamPlayerWrapper.this.isLive = isLive;
                                         StreamPlayerWrapper.this.duration = duration;
                                         seekType = SeekType.OSA_SEEK;

                                         Log.d(TAG, "Stream: " + stream + " is OSA_SEEK");
                                     }
                                 }

        );
    }

    @Override
    public void gotProblem(Problem problem) {
        Log.e(TAG, "Error! " + problem.getMessage() + "(" + problem.getErrorCode() + ")");

        if (!problem.isNormal() && listener != null) {
            listener.gotProblem(problem.getMessage());
//            CastApplication.getPlayerManager().gotProblem(problem.getMessage());
        }
    }

    public void requestPlay() {
        streamPlayer.requestPlay();
    }

    /**
     * This is a reference implementation for the implementation token authentication
     * see https://support.octoshape.com/entries/21368651-Implementing-Token-Authentication.
     *
     * @param authId ID provided by the @ OctoshapeSystem on start-up.
     */
    boolean authenticateStreamPlayer(String authId) throws IOException {

        final String ERROR = "ERROR_NO_AUTHID";
        URL url;

        // Simple Ticket
        if (authId == null)
            url = new URL("http://octoshape.com/broadcasters/webintegration/ticket/simple.php");

            // Stronger Security using Auth-ID
        else
            url = new URL(" http://octoshape.com/broadcasters/webintegration/ticket/authid.php?OctoshapeAuthid=" + authId);

        // Read server response
        BufferedReader in = new BufferedReader(new InputStreamReader(url.openStream()));
        String inputLine;
        ArrayList<String> response = new ArrayList<String>();
        while ((inputLine = in.readLine()) != null)
            response.add(inputLine);
        in.close();

        // Verifying server response and authenticate
        if (response.size() >= 2 && ERROR.equals(response.get(0))) {
            streamPlayer.setAuthorization(response.get(1), response.get(0));
            return true;
        } else {
            Log.e(TAG, "Error stream authentication. Invalid server response.");
            return false;
        }
    }

    void mediaPlayerStatusChanged(StreamManager.PlaybackState newState) {
        if (streamPlaybackState == null || streamPlaybackState != newState) {
            Log.d(TAG, "New playback state: " + newState + " for " + stream );
            streamPlaybackState = newState;

            notifyMediaPlayerListener(newState);
        } else {
            Log.d(TAG, "New playback state same as old state: " + newState + " for " + stream);
        }
    }

    void notifyMediaPlayerListener(StreamManager.PlaybackState newState) {
        Log.d(TAG, "Notify Media player listener. STATE: " + newState + ": " + stream);

        if (mediaPlayerListener == null) {
            Log.e(TAG, "No Media player listener associated for " + stream);
        } else {
            if (newState == StreamManager.PlaybackState.PLAYING) {
                mediaPlayerListener.onMediaPlaybackStarted();
            } else if (newState == StreamManager.PlaybackState.FINISHED) {
                mediaPlayerListener.onMediaPlaybackCompleted();
            }
        }
    }

    /**
     * Performs the seeking for OSA_SEEK streams, depending if it's live or not.
     * @param progress
     */
    public void seek(int progress) {
        if (seekType == SeekType.OSA_SEEK) {
            if (isLive) {
                streamPlayer.requestPlayLiveWithLatency(progress);
            } else {
                streamPlayer.requestPlayOndemandMediaTimePosition(progress);
            }
        }

    }

    public long getDuration() {
        return duration;
    }

}
